import passlib.utils.handlers as uh
from passlib.crypto.digest import compile_hmac
from passlib.utils.binary import h64

__all__: list[str] = []
_BNULL = b"\x00"


class sha1_crypt(  # type: ignore[misc]
    uh.HasManyBackends,
    uh.HasRounds,
    uh.HasSalt,
    uh.GenericHandler,
):
    """This class implements the SHA1-Crypt password hash, and follows the :ref:`password-hash-api`.

    It supports a variable-length salt, and a variable number of rounds.

    The :meth:`~passlib.ifc.PasswordHash.using` method accepts the following optional keywords:

    :type salt: str
    :param salt:
        Optional salt string.
        If not specified, an 8 character one will be autogenerated (this is recommended).
        If specified, it must be 0-64 characters, drawn from the regexp range ``[./0-9A-Za-z]``.

    :type salt_size: int
    :param salt_size:
        Optional number of bytes to use when autogenerating new salts.
        Defaults to 8 bytes, but can be any value between 0 and 64.

    :type rounds: int
    :param rounds:
        Optional number of rounds to use.
        Defaults to 480000, must be between 1 and 4294967295, inclusive.

    :type relaxed: bool
    :param relaxed:
        By default, providing an invalid value for one of the other
        keywords will result in a :exc:`ValueError`. If ``relaxed=True``,
        and the error can be corrected, a :exc:`~passlib.exc.PasslibHashWarning`
        will be issued instead. Correctable errors include ``rounds``
        that are too small or too large, and ``salt`` strings that are too long.

        .. versionadded:: 1.6
    """

    # --GenericHandler--
    name = "sha1_crypt"
    setting_kwds = ("salt", "salt_size", "rounds")
    ident = "$sha1$"
    checksum_size = 28
    checksum_chars = uh.HASH64_CHARS

    # --HasSalt--
    default_salt_size = 8
    max_salt_size = 64
    salt_chars = uh.HASH64_CHARS

    # --HasRounds--
    default_rounds = 480000  # current passlib default
    min_rounds = 1  # really, this should be higher.
    max_rounds = 4294967295  # 32-bit integer limit
    rounds_cost = "linear"

    @classmethod
    def from_string(cls, hash):
        rounds, salt, chk = uh.parse_mc3(hash, cls.ident, handler=cls)
        return cls(rounds=rounds, salt=salt, checksum=chk)

    def to_string(self, config=False):
        chk = None if config else self.checksum
        return uh.render_mc3(self.ident, self.rounds, self.salt, chk)

    backends = ("builtin",)

    # ---------------------------------------------------------------
    # builtin backend
    # ---------------------------------------------------------------
    @classmethod
    def _load_backend_builtin(cls):
        cls._set_calc_checksum_backend(cls._calc_checksum_builtin)
        return True

    def _calc_checksum_builtin(self, secret):
        if isinstance(secret, str):
            secret = secret.encode("utf-8")
        if _BNULL in secret:
            raise uh.exc.NullPasswordError(self)
        rounds = self.rounds
        # NOTE: this seed value is NOT the same as the config string
        result = (f"{self.salt}$sha1${rounds}").encode("ascii")
        # NOTE: this algorithm is essentially PBKDF1, modified to use HMAC.
        keyed_hmac = compile_hmac("sha1", secret)
        for _ in range(rounds):
            result = keyed_hmac(result)
        return h64.encode_transposed_bytes(result, self._chk_offsets).decode("ascii")

    _chk_offsets = [
        2,
        1,
        0,
        5,
        4,
        3,
        8,
        7,
        6,
        11,
        10,
        9,
        14,
        13,
        12,
        17,
        16,
        15,
        0,
        19,
        18,
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Protocol

if TYPE_CHECKING:
    from typing_extensions import Buffer, Self


class HashLike(Protocol):
    """Lifted from hashlib.pyi."""

    @property
    def digest_size(self) -> int: ...

    @property
    def block_size(self) -> int: ...

    @property
    def name(self) -> str: ...

    def copy(self) -> Self: ...

    def digest(self) -> bytes: ...

    def hexdigest(self) -> str: ...

    def update(self, data: Buffer, /) -> None: ...


SHAFunc = Callable[[], HashLike]



"""MD5-based algorithm used by Postgres for pg_shadow table"""

from hashlib import md5

import passlib.utils.handlers as uh
from passlib.utils import to_bytes

__all__ = [
    "postgres_md5",
]


class postgres_md5(uh.HasUserContext, uh.StaticHandler):
    """This class implements the Postgres MD5 Password hash, and follows the :ref:`password-hash-api`.

    It does a single round of hashing, and relies on the username as the salt.

    The :meth:`~passlib.ifc.PasswordHash.hash`, :meth:`~passlib.ifc.PasswordHash.genhash`, and :meth:`~passlib.ifc.PasswordHash.verify` methods all require the
    following additional contextual keywords:

    :type user: str
    :param user: name of postgres user account this password is associated with.
    """

    name = "postgres_md5"
    _hash_prefix = "md5"
    checksum_chars = uh.HEX_CHARS
    checksum_size = 32

    def _calc_checksum(self, secret):
        if isinstance(secret, str):
            secret = secret.encode("utf-8")
        user = to_bytes(self.user, "utf-8", param="user")
        return md5(secret + user).hexdigest()



import re
from base64 import b64decode, b64encode

import passlib.utils.handlers as uh
from passlib.crypto.digest import pbkdf1
from passlib.utils import to_unicode
from passlib.utils.compat import bascii_to_str

__all__ = [
    "fshp",
]


class fshp(  # type: ignore[misc]
    uh.HasRounds,
    uh.HasRawSalt,
    uh.HasRawChecksum,
    uh.GenericHandler,
):
    """This class implements the FSHP password hash, and follows the :ref:`password-hash-api`.

    It supports a variable-length salt, and a variable number of rounds.

    The :meth:`~passlib.ifc.PasswordHash.using` method accepts the following optional keywords:

    :param salt:
        Optional raw salt string.
        If not specified, one will be autogenerated (this is recommended).

    :param salt_size:
        Optional number of bytes to use when autogenerating new salts.
        Defaults to 16 bytes, but can be any non-negative value.

    :param rounds:
        Optional number of rounds to use.
        Defaults to 480000, must be between 1 and 4294967295, inclusive.

    :param variant:
        Optionally specifies variant of FSHP to use.

        * ``0`` - uses SHA-1 digest (deprecated).
        * ``1`` - uses SHA-2/256 digest (default).
        * ``2`` - uses SHA-2/384 digest.
        * ``3`` - uses SHA-2/512 digest.

    :type relaxed: bool
    :param relaxed:
        By default, providing an invalid value for one of the other
        keywords will result in a :exc:`ValueError`. If ``relaxed=True``,
        and the error can be corrected, a :exc:`~passlib.exc.PasslibHashWarning`
        will be issued instead. Correctable errors include ``rounds``
        that are too small or too large, and ``salt`` strings that are too long.

        .. versionadded:: 1.6
    """

    # --GenericHandler--
    name = "fshp"
    setting_kwds = ("salt", "salt_size", "rounds", "variant")
    checksum_chars = uh.PADDED_BASE64_CHARS
    ident = "{FSHP"
    # checksum_size is property() that depends on variant

    # --HasRawSalt--
    default_salt_size = 16  # current passlib default, FSHP uses 8
    max_salt_size = None

    # --HasRounds--
    # FIXME: should probably use different default rounds
    # based on the variant. setting for default variant (sha256) for now.
    default_rounds = 480000  # current passlib default, FSHP uses 4096
    min_rounds = 1  # set by FSHP
    max_rounds = 4294967295  # 32-bit integer limit - not set by FSHP
    rounds_cost = "linear"

    # --variants--
    default_variant = 1
    _variant_info = {
        # variant: (hash name, digest size)
        0: ("sha1", 20),
        1: ("sha256", 32),
        2: ("sha384", 48),
        3: ("sha512", 64),
    }
    _variant_aliases = dict(
        [(str(k), k) for k in _variant_info]
        + [(v[0], k) for k, v in _variant_info.items()]
    )

    @classmethod
    def using(cls, variant=None, **kwds):
        subcls = super().using(**kwds)
        if variant is not None:
            subcls.default_variant = cls._norm_variant(variant)
        return subcls

    variant = None

    def __init__(self, variant=None, **kwds):
        # NOTE: variant must be set first, since it controls checksum size, etc.
        self.use_defaults = kwds.get("use_defaults")  # load this early
        if variant is not None:
            variant = self._norm_variant(variant)
        elif self.use_defaults:
            variant = self.default_variant
            assert self._norm_variant(variant) == variant, (
                f"invalid default variant: {variant!r}"
            )
        else:
            raise TypeError("no variant specified")
        self.variant = variant
        super().__init__(**kwds)

    @classmethod
    def _norm_variant(cls, variant):
        if isinstance(variant, bytes):
            variant = variant.decode("ascii")
        if isinstance(variant, str):
            try:
                variant = cls._variant_aliases[variant]
            except KeyError:
                raise ValueError("invalid fshp variant")
        if not isinstance(variant, int):
            raise TypeError("fshp variant must be int or known alias")
        if variant not in cls._variant_info:
            raise ValueError("invalid fshp variant")
        return variant

    @property
    def checksum_alg(self):
        return self._variant_info[self.variant][0]

    @property
    def checksum_size(self):
        return self._variant_info[self.variant][1]

    _hash_regex = re.compile(
        r"""
            ^
            \{FSHP
            (\d+)\| # variant
            (\d+)\| # salt size
            (\d+)\} # rounds
            ([a-zA-Z0-9+/]+={0,3}) # digest
            $""",
        re.VERBOSE,
    )

    @classmethod
    def from_string(cls, hash):
        hash = to_unicode(hash, "ascii", "hash")
        m = cls._hash_regex.match(hash)
        if not m:
            raise uh.exc.InvalidHashError(cls)
        variant, salt_size, rounds, data = m.group(1, 2, 3, 4)
        variant = int(variant)
        salt_size = int(salt_size)
        rounds = int(rounds)
        try:
            data = b64decode(data.encode("ascii"))
        except TypeError:
            raise uh.exc.MalformedHashError(cls)
        salt = data[:salt_size]
        chk = data[salt_size:]
        return cls(salt=salt, checksum=chk, rounds=rounds, variant=variant)

    def to_string(self):
        chk = self.checksum
        salt = self.salt
        data = bascii_to_str(b64encode(salt + chk))
        return "{FSHP%d|%d|%d}%s" % (self.variant, len(salt), self.rounds, data)

    def _calc_checksum(self, secret):
        if isinstance(secret, str):
            secret = secret.encode("utf-8")
        # NOTE: for some reason, FSHP uses pbkdf1 with password & salt reversed.
        #       this has only a minimal impact on security,
        #       but it is worth noting this deviation.
        return pbkdf1(
            digest=self.checksum_alg,
            secret=self.salt,
            salt=secret,
            rounds=self.rounds,
            keylen=self.checksum_size,
        )



"""
passlib.hash - proxy object mapping hash scheme names -> handlers

==================
***** NOTICE *****
==================

This module does not actually contain any hashes. This file
is a stub that replaces itself with a proxy object.

This proxy object (passlib.registry._PasslibRegistryProxy)
handles lazy-loading hashes as they are requested.

The actual implementation of the various hashes is store elsewhere,
mainly in the submodules of the ``passlib.handlers`` subpackage.
"""


# XXX: if any platform has problem w/ lazy modules, could support 'non-lazy'
#      version which just imports all schemes known to list_crypt_handlers()

import sys

from passlib.registry import _proxy

sys.modules[__name__] = _proxy  # type: ignore[assignment]

# =============================================================================
# HACK: the following bit of code is unreachable, but it's presence seems to
#       help make autocomplete work for certain IDEs such as PyCharm.
#       this list is automatically regenerated using $SOURCE/admin/regen.py
# =============================================================================

# ----------------------------------------------------
# begin autocomplete hack (autogenerated 2016-11-10)
# ----------------------------------------------------
if False:
    from passlib.handlers.argon2 import argon2
    from passlib.handlers.bcrypt import bcrypt, bcrypt_sha256
    from passlib.handlers.cisco import cisco_asa, cisco_pix, cisco_type7
    from passlib.handlers.des_crypt import bigcrypt, bsdi_crypt, crypt16, des_crypt
    from passlib.handlers.digests import (
        hex_md4,
        hex_md5,
        hex_sha1,
        hex_sha256,
        hex_sha512,
        htdigest,
    )
    from passlib.handlers.django import (
        django_argon2,
        django_bcrypt,
        django_bcrypt_sha256,
        django_des_crypt,
        django_disabled,
        django_pbkdf2_sha1,
        django_pbkdf2_sha256,
        django_salted_md5,
        django_salted_sha1,
    )
    from passlib.handlers.fshp import fshp
    from passlib.handlers.ldap_digests import (  # type: ignore[attr-defined]
        ldap_bcrypt,
        ldap_bsdi_crypt,
        ldap_des_crypt,
        ldap_md5,
        ldap_md5_crypt,
        ldap_plaintext,
        ldap_salted_md5,
        ldap_salted_sha1,
        ldap_salted_sha256,
        ldap_salted_sha512,
        ldap_sha1,
        ldap_sha1_crypt,
        ldap_sha256_crypt,
        ldap_sha512_crypt,
    )
    from passlib.handlers.md5_crypt import apr_md5_crypt, md5_crypt
    from passlib.handlers.misc import plaintext, unix_disabled
    from passlib.handlers.mssql import mssql2000, mssql2005
    from passlib.handlers.mysql import mysql41, mysql323
    from passlib.handlers.oracle import oracle10, oracle11
    from passlib.handlers.pbkdf2 import (
        atlassian_pbkdf2_sha1,
        cta_pbkdf2_sha1,
        dlitz_pbkdf2_sha1,
        grub_pbkdf2_sha512,
        ldap_pbkdf2_sha1,
        ldap_pbkdf2_sha256,
        ldap_pbkdf2_sha512,
        pbkdf2_sha1,
        pbkdf2_sha256,
        pbkdf2_sha512,
    )
    from passlib.handlers.phpass import phpass
    from passlib.handlers.postgres import postgres_md5
    from passlib.handlers.roundup import ldap_hex_md5, ldap_hex_sha1, roundup_plaintext
    from passlib.handlers.scram import scram
    from passlib.handlers.scrypt import scrypt
    from passlib.handlers.sha1_crypt import sha1_crypt
    from passlib.handlers.sha2_crypt import sha256_crypt, sha512_crypt
    from passlib.handlers.sun_md5_crypt import sun_md5_crypt
    from passlib.handlers.windows import bsd_nthash, lmhash, msdcc, msdcc2, nthash
# ----------------------------------------------------
# end autocomplete hack
# ----------------------------------------------------



"""passlib.ext.django.models -- monkeypatch django hashing framework"""

from passlib.ext.django.utils import DjangoContextAdapter

# local
__all__ = ["password_context"]


#: adapter instance used to drive most of this
adapter = DjangoContextAdapter()

# the context object which this patches contrib.auth to use for password hashing.
# configuration controlled by ``settings.PASSLIB_CONFIG``.
password_context = adapter.context

#: hook callers should use if context is changed
context_changed = adapter.reset_hashers


# load config & install monkeypatch
adapter.load_model()



"""
passlib.utils.md4 - DEPRECATED MODULE, WILL BE REMOVED IN 2.0

MD4 should now be looked up through ``passlib.crypto.digest.lookup_hash("md4").const``,
which provides unified handling stdlib implementation (if present).
"""

from warnings import warn

from passlib.crypto.digest import lookup_hash

warn(
    "the module 'passlib.utils.md4' is deprecated as of Passlib 1.7, "
    "and will be removed in Passlib 2.0, please use "
    "'lookup_hash(\"md4\").const()' from 'passlib.crypto' instead",
    DeprecationWarning,
)

__all__ = ["md4"]

# this should use hashlib version if available,
# and fall back to builtin version.

md4 = lookup_hash("md4").const
del lookup_hash
